"""
.. module:: skrf.io.mdif

========================================
mdif (:mod:`skrf.io.mdif`)
========================================

Mdif class and utilities

.. autosummary::
   :toctree: generated/

   Mdif

"""
from __future__ import annotations

from itertools import product
from pathlib import Path
from typing import TextIO

import numpy as np

from ..frequency import Frequency
from ..mathFunctions import magdeg_2_reim
from ..network import Network, y2s, z2s
from ..networkSet import NetworkSet
from ..util import get_fid


class Mdif:
    """
    Class to read Generalized MDIF N-port files.

    Used to read the Generalized MDIF (GMDIF) file format [#]_ (.mdf extension).
    These files are used store Network parameters which vary with frequency
    and with one or more named parameters.

    Parameters
    ----------
    file : str or file-object
        mdif file to load

    References
    ----------
    .. [#] https://awrcorp.com/download/faq/english/docs/Users_Guide/data_file_formats.html#generalized_mdif

    Examples
    --------
    From filename

    >>> m = rf.Mdif('network.mdf')

    From file-object

    >>> file = open('network.mdf')
    >>> m = rf.Mdif(file)

    List of named parameters defined in the MDIF file

    >>> m.params

    Convert the MDIF to a NetworkSet

    >>> m.to_networkset()

    Using the data as a `NetworkSet` allows you to select a set of Networks
    from their named parameter values or interpolating between Networks.
    See `skrf.networkset.NetworkSet`.

    """

    def __init__(self, file: str | Path | TextIO):
        """
        Constructor

        Parameters
        ----------
        file : str, Path, or file-object
            mdif file to load
        """
        with get_fid(file) as fid:
            self.filename = fid.name
            self._comments = self._parse_comments(fid)
            self._params = []
            self._networks = self._parse_mdif(fid)

    @property
    def params(self) -> list:
        """
        Named parameters list.

        Returns
        -------
        list : list of string
            List of the named parameters defined in the MDIF file

        """
        return self._params

    @property
    def comments(self) -> list:
        """
        Comments defined in MDIF file.

        Returns
        -------
        list : list of string
            MDIF file Comments

        """
        return self._comments

    @property
    def networks(self) -> list:
        """
        List of Networks.

        Returns
        -------
        list : list of :class:`~skrf.network.Network`
            List of the Network stored in the MDIF file.

        """
        return self._networks

    def _parse_comments(self, fid) -> list:
        """
        Parse the comments stored in the MDIF file.

        Parameters
        ----------
        fid : file object

        Returns
        -------
        list : list of strings.

        """
        fid.seek(0)
        comments = []
        for line in fid:
            if line.strip().startswith('!'):
                comments.append(line[1:].strip())

            # Break out after the first network starts, thereby only
            # pulling out the comments for the entire NetworkSet, not
            # everything for each network. That's pulled out later.
            if line.strip().lower().startswith('begin'):
                break

        return comments


    def _parse_data(self, block_data: list) -> Network:
        """
        Parse the block of data corresponding to a set of parameters to a Network.

        Parameters
        ----------
        block_data : list of string
            block of data to parse

        Raises
        ------
        NotImplementedError
            if the data are not recognized

        Returns
        -------
        ntwk : rf.Network
            Network

        """
        kinds = []
        data_lines = []
        comments = []

        # Assumes default unit and format if not found during parsing
        frequency_unit = 'hz'
        parameter = None
        formt = 'ri'
        z0 = 50
        nb_lines_per_freq = 1
        ntwk_name = ''

        # Flag for adding comments to the network
        no_more_comments = False

        # Extract and group parameter informations and values
        for line in block_data:
            # Parse the option line (as in Touchstone)
            if line.startswith('#'):
                # Flag that no more comments are coming as we're now after the option line
                no_more_comments = True

                toks = line[1:].strip().split()
                # fill the option line with the missing defaults
                toks.extend(['ghz', 's', 'ma', 'r', '50'][len(toks):])
                frequency_unit = toks[0].lower()
                parameter = toks[1].lower()
                formt = toks[2].lower()
                z0 = toks[4]
                if frequency_unit not in ['hz', 'khz', 'mhz', 'ghz']:
                    raise NotImplementedError(f'ERROR: illegal frequency_unit {frequency_unit}',  )
                if parameter not in 'syzgh':
                    raise NotImplementedError(f'ERROR: illegal parameter value {parameter}')
                if formt not in ['ma', 'db', 'ri']:
                    raise NotImplementedError(f'ERROR: illegal format value {formt}')

            elif line.strip().startswith('!'):
                if line[1:].startswith(' network name:'):
                    ntwk_name = line.split(':')[-1].strip()

                # Append the comments to the list of comments for that data block
                if no_more_comments is not False:
                    comments.append(line.strip()[1:])

            # Parameter kinds (s11, z21, ...) are described as
            #
            # % kind1 kind2 ...
            # % ...
            # % ... kindN
            elif line.strip().lower().startswith('%'):
                kinds.append(line[1:].split())

            # Data are described as
            #
            # v_kind1 v_kind2 ...
            # ...
            # ... v_kindN
            elif line.strip():  # ignore lines with only whitespace
                data_lines.append(np.array(line.split(), dtype='float'))

        # grouping the data_lines in a single line for each frequency
        nb_lines_per_freq = len(kinds)
        data_lines = [np.concatenate(data_lines[idx*nb_lines_per_freq:idx*nb_lines_per_freq + nb_lines_per_freq])
                      for idx in range(int(len(data_lines)/nb_lines_per_freq))]
        kinds = [item for sublist in kinds for item in sublist]  # flattening
        kinds = [k.lower().replace('(complex)', '') for k in kinds]  # cleaning
        data = np.array(data_lines)
        f = data[:,0]

        # grouping two by two the columns convert data to complex arrays
        if formt == 'ri':
            values = data[:,1::2] + 1j*data[:,2::2]
        elif formt == 'ma':
            values = magdeg_2_reim(data[:,1::2], data[:,2::2])
        elif formt == 'db':
            values = ((10**(data[:,1::2]/20.0)) * np.exp(1j*np.pi/180 * data[:,2::2]))
        else:
            raise NotImplementedError('not implemented case')

        # Extracting s-parameters, depending of the file formatting
        # Nport in MDIF files exported by ADS
        if ('s[1,1]' in kinds):
            # deduce the rank (nb of ports) of the data
            rank = round(np.sqrt(sum('s[' in s for s in kinds)))
            s = np.zeros((len(f), rank, rank), dtype=complex)
            for m in range(rank):
                for n in range(rank):
                    s[:,m,n] = values[:, kinds.index(f's[{m+1},{n+1}]') - 1]

        # Nport as in AWR MDIF file-format description
        elif (parameter == 's') and all(k in kinds for k in ['n11x', 'n11y']):
            rank = round(np.sqrt(sum('n' in s for s in kinds)/2))
            s = values[:,:rank**2].reshape(len(f), rank, rank)
            # if rank is 2 and S21 before S12, swap S21 and S12
            if rank == 2 and (kinds.index('n21x') < kinds.index('n12x')):
                s[:, 1, 0], s[:, 0, 1] =  s[:, 0, 1].copy(), s[:, 1, 0].copy()

        # no S-parameter are found. Maybe Z-param instead?
        elif ('z[1,1]') in kinds:
            # deduce the rank (nb of ports) of the data
            rank = round(np.sqrt(sum('z[' in s for s in kinds)))
            z = np.zeros((len(f), rank, rank), dtype=complex)
            for m in range(rank):
                for n in range(rank):
                    z[:,m,n] = values[:, kinds.index(f'z[{m+1},{n+1}]') - 1]
            s = z2s(z, z0=z0)

        # no S nor Z-parameter are found. Maybe Y-param instead?
        elif ('y[1,1]') in kinds:
            # deduce the rank (nb of ports) of the data
            rank = round(np.sqrt(sum('y[' in s for s in kinds)))
            y = np.zeros((len(f), rank, rank), dtype=complex)
            for m in range(rank):
                for n in range(rank):
                    y[:,m,n] = values[:, kinds.index(f'y[{m+1},{n+1}]') - 1]
            s = y2s(z, z0=z0)

        else:
            raise NotImplementedError('Unrecognized case, probably not implemented')

        # building the Network
        freq = Frequency.from_f(f, unit=frequency_unit)
        ntwk = Network(frequency=freq, s=s, z0=z0, name=ntwk_name, comments="\n".join(comments))

        return ntwk

    def _parse_mdif(self, fid) -> list:
        """
        MDIF parser.

        Parameters
        ----------
        fid : file object

        Returns
        -------
        list: list of Networks
        """
        fid.seek(0)

        block_data = []
        ntwks = []
        params = dict()

        in_data_block = False

        for line in fid:

            # parse parameters:
            #
            # VAR param_name_1 = A
            # ...
            # VAR param_name_N = Z
            if line.lower().startswith('var'):
                # current parameter
                param_name, param_value = (s.strip() for s in line[3:].split('='))
                # remove the datatype "(blah)" in "varname(blah)" if any
                param_name = param_name.split('(')[0]
                # try to convert the value as a number
                self._params.append(param_name) if param_name not in self.params else self.params
                try:
                    params[param_name] = float(param_value)
                except ValueError:
                    params[param_name] = param_value.replace('"', '')

            # parse numerical data:
            #
            # begin XXX
            # ....
            # end
            if line.lower().startswith('end'):
                ntwk = self._parse_data(block_data)
                ntwk.params = params
                ntwks.append(ntwk)

                # reset parsed values
                in_data_block = False
                block_data = []
                params = dict()

            if in_data_block:
                block_data.append(line)

            if line.lower().startswith('begin'):
                in_data_block = True

        return ntwks

    def to_networkset(self) -> NetworkSet:
        """
        Return the MDIF data as a NetworkSet.

        Returns
        -------
        ns : :class:`~skrf.networkSet.NetworkSet`

        See Also
        --------
        from_networkset : Write a MDIF file from a NetworkSet.

        """
        ns = NetworkSet(self.networks)
        ns.comments = self.comments
        return ns

    @staticmethod
    def write(ns : NetworkSet,
              filename : str,
              values: dict | None = None,
              data_types: dict | None = None,
              comments: str | None = None,
              *,
              ads_compatible: bool = True,
              **kwargs):
        """
        Write a MDIF file from a NetworkSet.

        Parameters
        ----------
        ns : :class:`~skrf.networkSet.NetworkSet`
            NetworkSet to get the data from.
        filename : string
            Output MDIF file name.
        values : dictionary or None. Default is None.
            The keys of the dictionnary are MDIF variables and its values are
            a list of the parameter values.
            If None, then the values will be set to the NetworkSet names
            and the datatypes will be set to "string".
        data_types: dictionary or None. Default is None.
            The keys are MDIF variables and the value are datatypes
            specified by the following strings: "int", "double", and "string"
        comments: list of strings
            Comments to add to output_file.
            Each list items is a separate comment line
        ads_compatible: bool. Default is True.
            Indicates whether to write the file in a format that
            ADS will read properly.
        **kwargs: dictionary with extra arguments to pass through to the
            underlying write_touchstone() method in the Network class

        See Also
        --------
        io.mdif.Mdif : MDIF file class
        to_networkset : Return the MDIF data as a NetworkSet

        """

        if comments is None:
            comments = []
        if values is None:
            if ns.has_params():
                values = ns.params_values
            else:
                # using Network names as values
                v = list()
                for ntwk in ns:
                    v.append(ntwk.name)

                values = {"name": v}

        if data_types is None:
            if ns.has_params():
                data_types = ns.params_types
            else:
                # using Network names (->string)
                data_types = {"name": "string"}

        # Remove the return_string argument, as it's a required argument for this method
        kwargs.pop('return_string', None)

        # VAR datatypes
        dict_types = dict({"int": "0", "double": "1", "string": "2"})

        # open output_file
        with open(filename, "w") as mdif:

            # write comments
            for c in comments:
                mdif.write(f"! {c}\n")

            nports = ns[0].nports

            optionstring = Mdif.__create_optionstring(nports)

            for filenumber, ntwk in enumerate(ns):

                mdif.write("!" + "-" * 79 + "\n")

                for p in values:
                    # assign double as the datatype if none is specified
                    if p not in data_types:
                        data_types[p] = "double"

                    var_type = "" if ads_compatible else f"({dict_types[data_types[p]]})"
                    if data_types[p] == "string":
                        var_def_str = f'VAR {p}{var_type} = "{values[p][filenumber]}"'
                    else:
                        var_def_str = f"VAR {p}{var_type} = {values[p][filenumber]}"
                    mdif.write(var_def_str + "\n")

                mdif.write("\nBEGIN ACDATA\n")
                mdif.write(optionstring + "\n")
                mdif.write("! network name: " + ntwk.name + "\n")
                data = ntwk.write_touchstone(return_string=True, **kwargs)
                mdif.write(data)
                mdif.write("END\n\n")

    def __eq__(self, other) -> bool:
        """
        Test if two Mdif objects are equals.

        Test the equality between the NetworkSet under the hood.

        Parameters
        ----------
        other : :class:`~skrf.io.mdif.Mdif` object
            Mdif object to compare with.

        Returns
        -------
        bool

        """
        return self.to_networkset() == other.to_networkset()

    @staticmethod
    def __create_optionstring(nports):
        """create the options string based on the number of ports. Used in the Touchstone and MDIF formats"""

        if nports > 9:
            corestring = "n{}_{}x n{}_{}y "
        else:
            corestring = "n{}{}x n{}{}y "

        optionstring = "%F "

        if nports == 2:
            optionstring += "n11x n11y n21x n21y n12x n12y n22x n22y"

        else:
            # parse the option string for nports not equal to 2

            for i in product(list(range(1, nports + 1)), list(range(1, nports + 1))):

                optionstring += corestring.format(i[0], i[1], i[0], i[1])

                # special case for nports = 3
                if nports == 3:
                    # 3 ports
                    if not (np.remainder(i[1], 3)):
                        optionstring += "\n"

                # touchstone spec allows only 4 data pairs per line
                if nports >= 4:
                    if np.remainder(i[1], 4) == 0:
                        optionstring += "\n"
                    # NOTE: not sure if this is needed. Doesn't seem to be required by Microwave Office
                    if i[1] == nports:
                        optionstring += "\n"

        return optionstring



import os
import unittest

import numpy as np

import skrf as rf
from skrf.io import Touchstone, network_2_dataframe


class IOTestCase(unittest.TestCase):
    """
    """
    def setUp(self):
        """
        """
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'
        self.pickle_file = os.path.join(self.test_dir, 'pickled.p')
        self.hfss_oneport_file = os.path.join(self.test_dir, 'hfss_oneport.s1p')
        self.hfss_twoport_file = os.path.join(self.test_dir, 'hfss_twoport.s2p')
        self.ntwk1 = rf.Network(os.path.join(self.test_dir, 'ntwk1.s2p'))
        self.ntwk2 = rf.Network(os.path.join(self.test_dir, 'ntwk2.s2p'))
        self.ntwk3 = rf.Network(os.path.join(self.test_dir, 'ntwk3.s2p'))
        self.short = rf.Network(os.path.join(self.test_dir, 'short.s1p'))
        self.match = rf.Network(os.path.join(self.test_dir, 'match.s1p'))
        self.open = rf.Network(os.path.join(self.test_dir, 'open.s1p'))
        self.ntwk_comments_file = os.path.join(self.test_dir, 'comments.s3p')
        self.test_files = [os.path.join(self.test_dir, test_file) for test_file in ['ntwk1.s2p', 'ntwk2.s2p']]
        self.embeding_network= rf.Network(os.path.join(self.test_dir, 'embedingNetwork.s2p'))
        self.freq = rf.F(75, 110, 101, unit='GHz')

    def read_write(self,obj):
        """
        function to test write/read equivalence for an obj which has
        __eq__ defined
        """
        rf.write(self.pickle_file,obj)
        self.assertEqual(rf.read(self.pickle_file), obj)
       # os.remove(self.pickle_file)



    def test_read_all(self):
        rf.read_all(self.test_dir)

    def test_read_all_files(self):
        rf.read_all(files=self.test_files)

    def test_save_sesh(self):
        a=self.ntwk1
        b=self.ntwk2
        c=self.ntwk3
        rf.save_sesh(locals(),self.pickle_file )
        #os.remove(self.pickle_file)

    def test_write_all_dict(self):
        d = dict(a=self.ntwk1, b=self.ntwk2,   c=self.ntwk3)
        rf.write_all(d, dir =self.test_dir )
        os.remove(os.path.join(self.test_dir, 'a.ntwk'))
        os.remove(os.path.join(self.test_dir, 'b.ntwk'))
        os.remove(os.path.join(self.test_dir, 'c.ntwk'))

    def test_readwrite_network(self):
        self.read_write(self.ntwk1)

    def test_readwrite_list_of_network(self):
        self.read_write([self.ntwk1, self.ntwk2])

    def test_readwrite_networkSet(self):
        """
        test_readwrite_networkSet
        TODO: need __eq__ method for NetworkSet
        This doesnt test equality between  read/write, because there is no
        __eq__ test for NetworkSet. it only tests for other errors
        """
        rf.write(self.pickle_file,rf.NS([self.ntwk1, self.ntwk2]))
        rf.read(self.pickle_file)
        #self.assertEqual(rf.read(self.pickle_file), rf.NS([self.ntwk1, self.ntwk2])
        #os.remove(self.pickle_file)

    def test_readwrite_frequency(self):
        freq = rf.Frequency(1,10,10,'ghz')
        self.read_write(freq)

    def test_readwrite_calibration(self):
        ideals, measured = [], []
        std_list = [self.short, self.match,self.open]

        for ntwk in std_list:
            ideals.append(ntwk)
            measured.append(self.embeding_network ** ntwk)

        cal = rf.Calibration(\
                ideals = ideals,\
                measured = measured,\
                type = 'one port',\
                is_reciprocal = True,\
                )

        original = cal
        rf.write(self.pickle_file, original)
        unpickled = rf.read(self.pickle_file)
        # TODO: this test should be more extensive
        self.assertEqual(original.ideals, unpickled.ideals)
        self.assertEqual(original.measured, unpickled.measured)

        os.remove(self.pickle_file)

    def test_readwrite_media(self):
        a_media = rf.media.DefinedGammaZ0(
            frequency = self.freq,
            gamma = 1j*np.ones(101) ,
            z0 =  50*np.ones(101),
            )
        self.read_write(a_media)
    @unittest.skip
    def test_readwrite_media_func_propgamma(self):
        a_media = rf.media.Media(
            frequency = self.freq,
            propagation_constant = lambda :1j ,
            characteristic_impedance =  lambda :50,
            )
        self.read_write(a_media)
    @unittest.skip
    def test_readwrite_RectangularWaveguide(self):
        a_media = rf.media.RectangularWaveguide(
            frequency = self.freq,
            a=100*rf.mil,
            z0=50,
            )
        self.read_write(a_media)
    @unittest.skip
    def test_readwrite_DistributedCircuit(self):
        one = np.ones(self.freq.npoints)
        a_media = rf.media.DistributedCircuit(
            frequency = self.freq,
            R=1e5*one, G=1*one, I=1e-6*one, C=8e-12*one
            )
        self.read_write(a_media)
    @unittest.skip
    def test_readwrite_Freespace(self):
        a_media = rf.media.Freespace(
            frequency = self.freq,
            )
        self.read_write(a_media)

    def test_snp_json_roundtrip(self):
        """
        Tests if snp object saved to json and reloaded is still the same.
        """
        given = self.ntwk1
        actual = rf.from_json_string(rf.to_json_string(given))
        self.assertEqual(actual, given)
        self.assertEqual(actual.frequency, given.frequency)
        self.assertEqual(actual.name, given.name)
        self.assertEqual(actual.comments, given.comments)
        self.assertEqual(actual.z0.tolist(), given.z0.tolist())
        self.assertEqual(actual.port_names, given.port_names)
        self.assertEqual(actual.variables, given.variables)

    def test_touchstone_get_comment_variables(self):
        """
        Tests if comments are parsed correctly with get_comment_variables() method.
        """

        given = {'p1': ('.03', ''), 'p2': ('0.03', ''), 'p3': ('100', ''), 'p4': ('2.5', 'um')}
        actual = Touchstone(self.ntwk_comments_file).get_comment_variables()
        self.assertEqual(given, actual)

    def test_network_2_dataframe_equal(self):
        df_method = self.ntwk1.to_dataframe()
        df_function = network_2_dataframe(self.ntwk1)

        assert df_method.equals(df_function)

    def test_network_2_dataframe_columns(self):
        s = np.random.default_rng().standard_normal((1, 11, 11))
        f = [1]
        netw = rf.Network(s=s, f=f)

        df = netw.to_dataframe()
        assert s.size == df.values.size
        assert "s_db 1_11" in df.columns
        assert "s_db 11_1" in df.columns

    def test_network_2_dataframe_port_sep(self):
        for port_sep in ["", "_", ","]:
            df = self.ntwk1.to_dataframe(port_sep=port_sep)

            assert len(df.columns == self.ntwk1.nports ** 2)
            assert f"s_db 2{port_sep}1" in df.columns

    def test_network_2_dataframe_port_sep_auto(self):
        f = [1]
        for ports in [1, 2, 4, 8, 10, 11, 16]:
            s = np.random.default_rng().standard_normal((1, ports, ports))
            netw = rf.Network(s=s, f=f)

            df = netw.to_dataframe()

            if ports <= 10:
                assert "s_db 11" in df.columns
            else:
                assert "s_db 1_1" in df.columns



"""
this test on tests ability for all media class to pass construction
of all general circuit components

"""
import unittest

import numpy as np
from scipy.constants import mil

import skrf as rf
from skrf.media import CPW, CircularWaveguide, Coaxial, DistributedCircuit, Freespace, MLine, RectangularWaveguide


class MediaTestCase:
    """Base class, contains tests for all media."""
    def test_gamma(self):
        self.media.gamma

    def test_z0_value(self):
        self.media.z0

    def test_match(self):
        self.media.match()

    def test_load(self):
        self.media.load(1)

    def test_short(self):
        self.media.short()

    def test_open(self):
        self.media.open()

    def test_capacitor(self):
        self.media.capacitor(1)

    def test_capacitor_q(self):
        self.media.capacitor_q(1, 2, 3)

    def test_inductor(self):
        self.media.inductor(1)

    def test_inductor_q(self):
        self.media.capacitor_q(1, 2, 3)

    def test_impedance_mismatch(self):
        self.media.impedance_mismatch(1, 2)

    def test_tee(self):
        self.media.tee()

    def test_splitter(self):
        self.media.splitter(4)

    def test_thru(self):
        self.media.thru()

    def test_line(self):
        self.media.line(1)

    def test_line_floating(self):
        self.media.line_floating(1)

    def test_delay_load(self):
        self.media.delay_load(1,2)

    def test_delay_short(self):
        self.media.delay_short(1)

    def test_delay_open(self):
        self.media.delay_open(1)

    def test_shunt_delay_load(self):
        self.media.shunt_delay_load(1,1)

    def test_shunt_delay_short(self):
        self.media.shunt_delay_short(1)

    def test_shunt_delay_open(self):
        self.media.shunt_delay_open(1)

    def test_shunt_capacitor(self):
        self.media.shunt_capacitor(1)

    def test_shunt_inductor(self):
        self.media.shunt_inductor(1)

    def test_Z0_deprecation_warning(self):
        with self.assertWarns(DeprecationWarning) as context:
            self.media.Z0

    def test_embed_deprecation_warning(self):
        with self.assertWarns(DeprecationWarning) as context:
            self.media.line(1, unit = 'deg', embed = True)

class Z0InitDeprecationTestCase(unittest.TestCase):
    def setUp(self):
        self.frequency = rf.Frequency(1,2,2,'GHz')

    def testZ0InitDeprecation(self):
        # 1-conductor waveguide media
        with self.assertWarns(DeprecationWarning) as context:
            cw = CircularWaveguide(self.frequency, z0 = 50)
            self.assertTrue(np.all(cw.z0 == 50))
        with self.assertWarns(DeprecationWarning) as context:
            rw = RectangularWaveguide(self.frequency, z0 = 50)
            self.assertTrue(np.all(rw.z0 == 50))
        # 2-conductor other media
        with self.assertWarns(DeprecationWarning) as context:
            coax = Coaxial(self.frequency, z0 = 50)
            self.assertTrue(np.all(coax.z0 == 50))
        with self.assertWarns(DeprecationWarning) as context:
            cpw = CPW(self.frequency, z0 = 50)
            self.assertTrue(np.all(cpw.z0 == 50))
        with self.assertWarns(DeprecationWarning) as context:
            air = Freespace(self.frequency, z0 = 50)
            self.assertTrue(np.all(air.z0 == 50))
        with self.assertWarns(DeprecationWarning) as context:
            mlin = MLine(self.frequency, z0 = 50)
            self.assertTrue(np.all(mlin.z0 == 50))


class FreespaceTestCase(MediaTestCase, unittest.TestCase):
    def setUp(self):
        self.frequency = rf.Frequency(75,110,101,'GHz')
        self.media = Freespace(self.frequency)

    def test_z0_value(self):
        self.assertEqual(round(\
            self.media.z0[0].real), 377)


class CPWTestCase(MediaTestCase, unittest.TestCase):
    def setUp(self):
        self.frequency = rf.Frequency(75,110,101,'GHz')
        self.media = CPW(\
            frequency=self.frequency,
            w=10e-6,
            s=5e-6,
            ep_r=11.7,
            t=1e-6,
            rho=22e-9)


class RectangularWaveguideTestCase(MediaTestCase, unittest.TestCase):
    def setUp(self):
        self.frequency = rf.Frequency(75,110,101,'GHz')
        self.media = RectangularWaveguide(\
            frequency=self.frequency,
            a=100*mil,
            )


class DistributedCircuitTestCase(MediaTestCase, unittest.TestCase):
    def setUp(self):
        self.frequency = rf.Frequency(75,110,101,'GHz')
        self.media = DistributedCircuit(\
            frequency=self.frequency,
            L=1,C=1,R=0,G=0
            )


suite = unittest.TestSuite()
loader = unittest.TestLoader()


suite.addTests([\
    loader.loadTestsFromTestCase(FreespaceTestCase),
    loader.loadTestsFromTestCase(CPWTestCase),
    loader.loadTestsFromTestCase(RectangularWaveguideTestCase),
    loader.loadTestsFromTestCase(DistributedCircuitTestCase),
    loader.loadTestsFromTestCase(Z0InitDeprecationTestCase),
    ])

#suite = unittest.TestLoader().loadTestsFromTestCase(FreespaceTestCase)
unittest.TextTestRunner(verbosity=2).run(suite)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Sequence

import functools
from enum import Enum

import numpy as np
import pyvisa

import skrf
from skrf.vi import vna


class OP(bytes, Enum):
    NOP = b"\x00"
    INDICATE = b"\x0d"
    READ = b"\x10"
    READ2 = b"\x11"
    READ4 = b"\x12"
    READFIFO = b"\x18"
    WRITE = b"\x20"
    WRITE2 = b"\x21"
    WRITE4 = b"\x22"
    WRITE8 = b"\x23"
    WRITEFIFO = b"\x28"


class REG_ADDR(bytes, Enum):
    SWEEP_START = b"\x00"
    SWEEP_STEP = b"\x10"
    SWEEP_POINTS = b"\x20"
    VALS_PER_FREQ = b"\x22"
    RAW_SAMPLES_MODE = b"\x26"
    VALS_FIFO = b"\x30"
    DEVICE_VARIANT = b"\xf0"
    PROTOCOL_VERSION = b"\xf1"
    HARDWARE_REV = b"\xf2"
    FIRMWARE_MAJOR = b"\xf3"
    FIRMWARE_MINOR = b"\xf4"


class NanoVNAv2(vna.VNA):
    """NanoVNAv2.

    This class connects to the NanoVNA V2 using a binary protocol over USB.
    It should also be compatible with other devices, if they use the same `protocol <https://nanorfe.com/nanovna-v2-user-manual.html#__RefHeading___Toc2537_2953165397>`_.
    Some variants of the NanoVNA use a text protocol and are not supported.

    .. warning::

        The NanoVNA V2 only returns uncalibrated networks, regardless of the calibration on the device itself.
        See section `Examples`_.

    Notes
    -----

    Tested devices:

    * `NanoVNA V2 <https://nanorfe.com/de/nanovna-v2.html>`_
    * `LiteVNA <https://www.zeenko.tech/litevna>`_

    .. _Examples:
    Examples
    --------

    Basic S11 and S21 measurement:

    .. code-block:: python

        import skrf
        from skrf.vi.vna.nanovna import NanoVNAv2

        vna = NanoVNAv2("ASRL/dev/ttyACM0::INSTR")  # Linux
        # vna = NanoVNAv2("ASRL1::INSTR")  # Windows
        freq = skrf.Frequency(start=1, stop=2, unit='GHz', npoints=101)
        s11, s21 = vna.get_s11_s21()

    1-port calibration assuming ideal reference standards:

    .. code-block:: python

        import skrf

        # load uncalibrated measurement
        raw_s1p = skrf.Network("measurement.s1p")

        # load s1p networks of measured standards
        cal_measured = [skrf.Network("short.s1p"), skrf.Network("open.s1p"), skrf.Network("load.s1p"]

        # get ideal standards
        line = skrf.DefinedGammaZ0(frequency=cal_measured[0].frequency, z0=50)
        cal_ideals = [line.short(nports=1), line.open(nports=1), line.match(nports=1)]

        # run calibration
        cal = skrf.OnePort(ideals=cal_ideals, measured=cal_measured)
        cal.run()

        # apply calibration
        calibrated_s1p = cal.apply_cal(raw_s1p)
    """

    _scpi = False

    def __init__(self, address, backend: str = "@py"):
        super().__init__(address, backend)
        if not isinstance(self._resource, pyvisa.resources.SerialInstrument):
            raise RuntimeError(
                "NanoVNAv2 can only be a serial instrument. "
                f"{address} yields a {self._resource.__class__.__name__}"
            )

        self.read_bytes = self._resource.read_bytes
        self.write_raw = self._resource.write_raw

        self._reset_protocol()

        self.frequency = skrf.Frequency(start=1e6, stop=10e6, npoints=201)

    def _reset_protocol(self):
        self.write_raw(b"\x00\x00\x00\x00\x00\x00\x00\x00")

    def query(self, op: OP, addr: REG_ADDR | bytes, nbytes: int) -> bytes:
        cmd = op + addr
        self.write_raw(cmd)
        return self.read_bytes(nbytes)

    def write(self, op: OP, addr: REG_ADDR | bytes, nbytes: int, arg: int) -> None:
        arg = int(arg).to_bytes(nbytes, byteorder="little", signed=False)

        if op == OP.WRITEFIFO:
            cmd = op + addr + nbytes.to_bytes(1) + arg
        else:
            cmd = op + addr + arg

        self.write_raw(cmd)

    @property
    def id(self) -> str:
        var = self.query(OP.READ, REG_ADDR.DEVICE_VARIANT, 1)
        var = int.from_bytes(var, 'little')
        return str(var)

    @property
    def device_info(self) -> str:
        variant = self.id
        protocol = self.query(OP.READ, REG_ADDR.PROTOCOL_VERSION, 1)
        protocol = int.from_bytes(protocol, 'little')
        hardware = self.query(OP.READ, REG_ADDR.HARDWARE_REV, 1)
        hardware = int.from_bytes(hardware, 'little')
        fw_major = self.query(OP.READ, REG_ADDR.FIRMWARE_MAJOR, 1)
        fw_major = int.from_bytes(fw_major, 'little')
        fw_minor = self.query(OP.READ, REG_ADDR.FIRMWARE_MAJOR, 1)
        fw_minor = int.from_bytes(fw_minor, 'little')

        return (
            f"NanoVNAv2\n"
            f"\tVariant:{variant}\n"
            f"\tProtocol Version:{protocol}\n"
            f"\tHardware Version: {hardware}\n"
            f"\tFirmware Version: {fw_major}.{fw_minor}"
        )


    @property
    def freq_start(self) -> float:
        return self._freq.start

    @freq_start.setter
    def freq_start(self, f: int) -> None:
        self.write(OP.WRITE8, REG_ADDR.SWEEP_START, 8, f)
        self._freq = skrf.Frequency(start=f, stop=self._freq.stop, npoints=self._freq.npoints)

    @property
    def freq_stop(self) -> float:
        return self._freq.stop

    @freq_stop.setter
    def freq_stop(self, f: int) -> None:
        self._freq = skrf.Frequency(start=self._freq.start, stop=f, npoints=self._freq.npoints)
        self.write(OP.WRITE8, REG_ADDR.SWEEP_STEP, 8, self._freq.step)

    @property
    def freq_step(self) -> float:
        return self._freq.step

    @freq_step.setter
    def freq_step(self, f: int) -> None:
        npoints = (self._freq.stop - self._freq.start + f) / f
        npoints = int(npoints.round())
        self._freq = skrf.Frequency(start=self._freq.start, stop=self._freq.stop, npoints=npoints)
        self.write(OP.WRITE2, REG_ADDR.SWEEP_POINTS, 2, npoints)

    @property
    def npoints(self) -> int:
        return self._freq.npoints

    @npoints.setter
    def npoints(self, n: int) -> None:
        self.write(OP.WRITE2, REG_ADDR.SWEEP_POINTS, 2, n)
        self._freq = skrf.Frequency(start=self._freq.start, stop=self._freq.stop, npoints=n)

    @property
    def frequency(self) -> skrf.Frequency:
        return self._freq

    @frequency.setter
    def frequency(self, f: skrf.Frequency):
        self.write(OP.WRITE8, REG_ADDR.SWEEP_START, 8, f.start)
        self.write(OP.WRITE8, REG_ADDR.SWEEP_STEP, 8, f.step)
        self.write(OP.WRITE2, REG_ADDR.SWEEP_POINTS, 2, f.npoints)
        self._freq = f

    def clear_fifo(self) -> None:
        self.write(OP.WRITE, REG_ADDR.VALS_FIFO, 1, 0)

    def _convert_bytes_to_sparams(n: int, raw: bytearray) -> tuple[np.ndarray, np.ndarray]:
        s11 = np.zeros(n, dtype=complex)
        s21 = np.zeros_like(s11)

        from_bytes = functools.partial(int.from_bytes, byteorder='little', signed=True)

        for i in range(n):
            start = i * 32
            stop = (i+1) * 32
            chunk = raw[start:stop]

            fwd0re = from_bytes(chunk[0:4])
            fwd0im = from_bytes(chunk[4:8])
            rev0re = from_bytes(chunk[8:12])
            rev0im = from_bytes(chunk[12:16])
            rev1re = from_bytes(chunk[16:20])
            rev1im = from_bytes(chunk[20:24])
            freqIndex = from_bytes(chunk[24:26])

            a1 = complex(fwd0re, fwd0im)
            b1 = complex(rev0re, rev0im)
            b2 = complex(rev1re, rev1im)

            s11[freqIndex] = b1 / a1
            s21[freqIndex] = b2 / a1

        return s11, s21

    def get_s11_s21(self) -> tuple[skrf.Network, skrf.Network]:
        n = self._freq.npoints
        self.clear_fifo()

        raw = bytearray()
        n_remaining = n

        while n_remaining > 0:
            len_segment = 255 if n_remaining > 255 else n_remaining
            n_remaining = n_remaining - len_segment
            self.write(OP.READFIFO, REG_ADDR.VALS_FIFO, 1, len_segment)
            raw.extend(self.read_bytes(32 * len_segment))

        s11, s21 = skrf.Network(), skrf.Network()
        s11.frequency = self._freq.copy()
        s21.frequency = self._freq.copy()

        s11.s, s21.s = NanoVNAv2._convert_bytes_to_sparams(n, raw)

        return s11, s21

    def get_sdata(self, a: int = 1, b: int = 1) -> skrf.Network:
        """
        Get S-parameter as 1-port :class:`skrf.Network`.

        Parameters
        ----------
        a:
            Input port of VNA. Can be 1 or 2, default is 1.
        b:
            Output port of VNA. Can only be 1, default is 1.

        Returns
        -------
        :class:`skrf.Network`
            Measured S-parameter
        """
        if a not in (1, 2) or b != 1:
            raise RuntimeError(
                "NanoVNA V2 can only measure S11 and S21"
            )

        return self.get_s11_s21()[a-1]

    def get_snp_network(self, ports: Sequence[int | None]) -> skrf.Network:
        """
        Get custom n-port :class:`skrf.Network`.

        Parameters
        ----------
        ports:
            Port mapping. Can be any sequence of 1, 2 and None

        Returns
        -------
        :class:`skrf.Network`
            Measured S-parameters in custom network
        """
        nwk_s11, nwk_s21 = self.get_s11_s21()
        frequency = self._freq.copy()

        s_parameters = np.zeros((len(frequency), len(ports), len(ports)), dtype=complex)
        snp_network = skrf.Network(frequency=frequency, s=s_parameters)

        for i, i_port in enumerate(ports):
            if i_port is not None:
                for j, j_port in enumerate(ports):
                    if i_port == 1 and j_port == 1:
                        snp_network.s[:, i, j] = nwk_s11.s[:, 0, 0]
                    elif i_port == 2 and j_port == 1:
                        snp_network.s[:, i, j] = nwk_s21.s[:, 0, 0]

        return snp_network



import os
import unittest

import numpy as np
import pytest
from scipy.constants import c, pi

import skrf as rf
from skrf.media import MLine


class DeembeddingTestCase(unittest.TestCase):
    """
    Testcase for the Deembedding class
    Pseudo-netlists for s-parameter files used in these tests
    For open-short de-embedding:
    - deemb_ind.s2p
        P1      (1 0) port
        Cpad1   (1 0) capacitor c=25fF
        Rline1  (1 2) resistor r=2ohm
        Dut_ind (2 3) inductor l=1nH
        Rline2  (3 4) resistor r=2ohm
        Cpad2   (4 0) capacitor c=25fF
        Cp2p    (1 4) capacitor c=10fF
        P2      (4 0) port
    - deemb_open.s2p
        P1      (1 0) port
        Cpad1   (1 0) capacitor c=25fF
        Rline1  (1 2) resistor r=2ohm
        Rline2  (3 4) resistor r=2ohm
        Cpad2   (4 0) capacitor c=25fF
        Cp2p    (1 4) capacitor c=10fF
        P2      (4 0) port
    - deemb_short.s2p
        P1      (1 0) port
        Cpad1   (1 0) capacitor c=25fF
        Rline1  (1 0) resistor r=2ohm
        Rline2  (0 4) resistor r=2ohm
        Cpad2   (4 0) capacitor c=25fF
        Cp2p    (1 4) capacitor c=10fF
        P2      (4 0) port
    For open de-embedding:
    - deemb_ind7.s2p
        P1      (1 0) port
        Cpad1   (1 0) capacitor c=25fF
        Dut_ind (1 2) inductor l=1nH
        Cpad2   (2 0) capacitor c=25fF
        Cp2p    (1 2) capacitor c=10fF
        P2      (2 0) port
    - deemb_open7.s2p
        P1      (1 0) port
        Cpad1   (1 0) capacitor c=25fF
        Cpad2   (2 0) capacitor c=25fF
        Cp2p    (1 2) capacitor c=10fF
        P2      (2 0) port
    For short-open de-embedding:
    - deemb_ind2.s2p
        P1      (1 0) port
        Rline1  (1 2) resistor r=2ohm
        Cpad1   (2 0) capacitor c=25fF
        Dut_ind (2 3) inductor l=1nH
        Cpad2   (3 0) capacitor c=25fF
        Cp2p    (2 3) capacitor c=10fF
        Rline2  (3 4) resistor r=2ohm
        P2      (4 0) port
    - deemb_open2.s2p
        P1      (1 0) port
        Rline1  (1 2) resistor r=2ohm
        Cpad1   (2 0) capacitor c=25fF
        Cpad2   (3 0) capacitor c=25fF
        Cp2p    (2 3) capacitor c=10fF
        Rline2  (3 4) resistor r=2ohm
        P2      (4 0) port
    - deemb_short2.s2p
        P1      (1 0) port
        Rline1  (1 0) resistor r=2ohm
        Cpad1   (0 0) capacitor c=25fF
        Cpad2   (0 0) capacitor c=25fF
        Cp2p    (0 0) capacitor c=10fF
        Rline2  (0 4) resistor r=2ohm
        P2      (4 0) port
    For short de-embedding:
    - deemb_ind8.s2p
        P1      (1 0) port
        Rline1  (1 2) resistor r=2ohm
        Dut_ind (2 3) inductor l=1nH
        Rline2  (3 4) resistor r=2ohm
        P2      (4 0) port
    - deemb_short8.s2p
        P1      (1 0) port
        Rline1  (1 0) resistor r=2ohm
        Rline2  (0 4) resistor r=2ohm
        P2      (4 0) port
    For SplitPi de-embedding:
    - deemb_ind3.s2p
        P1      (1 0) port
        Cpad1   (1 0) capacitor c=25fF
        Rline1  (1 2) resistor r=2ohm
        Dut_ind (2 3) inductor l=1nH
        Rline2  (3 4) resistor r=2ohm
        Cpad2   (4 0) capacitor c=25fF
        P2      (4 0) port
    - deemb_thru3.s2p
        P1      (1 0) port
        Cpad1   (1 0) capacitor c=25fF
        Rline1  (1 2) resistor r=2ohm
        Rline2  (3 4) resistor r=2ohm
        Cpad2   (4 0) capacitor c=25fF
        P2      (4 0) port
    For SplitTee de-embedding:
    - deemb_ind4.s2p
        P1      (1 0) port
        Rline1  (1 2) resistor r=2ohm
        Cpad1   (2 0) capacitor c=25fF
        Dut_ind (2 3) inductor l=1nH
        Cpad2   (3 0) capacitor c=25fF
        Rline2  (3 4) resistor r=2ohm
        P2      (4 0) port
    - deemb_thru4.s2p
        P1      (1 0) port
        Rline1  (1 2) resistor r=2ohm
        Cpad1   (2 0) capacitor c=25fF
        Cpad2   (3 0) capacitor c=25fF
        Rline2  (3 4) resistor r=2ohm
        P2      (4 0) port
    For AdmittanceCancel de-embedding:
    - deemb_ind5.s2p
        P1      (1 0) port
        Cpad1   (1 0) capacitor c=25fF
        Dut_ind (1 2) inductor l=1nH
        Cpad2   (2 0) capacitor c=25fF
        P2      (2 0) port
    - deemb_thru5.s2p
        P1      (1 0) port
        Cpad1   (1 0) capacitor c=25fF
        Cpad2   (1 0) capacitor c=25fF
        P2      (1 0) port
    For ImpedanceCancel de-embedding:
    - deemb_ind6.s2p
        P1      (1 0) port
        Rline1  (1 2) resistor r=2ohm
        Dut_ind (2 3) inductor l=1nH
        Rline2  (3 4) resistor r=2ohm
        P2      (4 0) port
    - deemb_thru6.s2p
        P1      (1 0) port
        Rline1  (1 2) resistor r=2ohm
        Rline2  (2 3) resistor r=2ohm
        P2      (3 0) port
    """

    def setUp(self):
        """
        Read in all the network data required for tests
        """
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/deembed/'

        # for open-short testing
        self.raw = rf.Network(os.path.join(self.test_dir, 'deemb_ind.s2p'))
        self.open = rf.Network(os.path.join(self.test_dir, 'deemb_open.s2p'))
        self.short = rf.Network(os.path.join(self.test_dir, 'deemb_short.s2p'))

        # for spot frequency checking
        self.raw_1f = self.raw['10GHz']
        self.open_1f = self.open['10GHz']
        self.short_1f = self.short['10GHz']

        # for open testing
        self.raw7 = rf.Network(os.path.join(self.test_dir, 'deemb_ind7.s2p'))
        self.open7 = rf.Network(os.path.join(self.test_dir, 'deemb_open7.s2p'))

        # for spot frequency checking
        self.raw7_1f = self.raw7['10GHz']
        self.open7_1f = self.open7['10GHz']

        # for short-open testing
        self.raw2 = rf.Network(os.path.join(self.test_dir, 'deemb_ind2.s2p'))
        self.open2 = rf.Network(os.path.join(self.test_dir, 'deemb_open2.s2p'))
        self.short2 = rf.Network(os.path.join(self.test_dir, 'deemb_short2.s2p'))

        # for spot frequency checking
        self.raw2_1f = self.raw2['10GHz']
        self.open2_1f = self.open2['10GHz']
        self.short2_1f = self.short2['10GHz']

        # for short testing
        self.raw8 = rf.Network(os.path.join(self.test_dir, 'deemb_ind8.s2p'))
        self.short8 = rf.Network(os.path.join(self.test_dir, 'deemb_short8.s2p'))

        # for spot frequency checking
        self.raw8_1f = self.raw8['10GHz']
        self.short8_1f = self.short8['10GHz']

        # for spilit-pi testing
        self.raw3 = rf.Network(os.path.join(self.test_dir, 'deemb_ind3.s2p'))
        self.thru3 = rf.Network(os.path.join(self.test_dir, 'deemb_thru3.s2p'))

        # for spot frequency checking
        self.raw3_1f = self.raw3['10GHz']
        self.thru3_1f = self.thru3['10GHz']

        # for spilit-tee testing
        self.raw4 = rf.Network(os.path.join(self.test_dir, 'deemb_ind4.s2p'))
        self.thru4 = rf.Network(os.path.join(self.test_dir, 'deemb_thru4.s2p'))

        # for spot frequency checking
        self.raw4_1f = self.raw4['10GHz']
        self.thru4_1f = self.thru4['10GHz']

        # for admittance cancel testing
        self.raw5 = rf.Network(os.path.join(self.test_dir, 'deemb_ind5.s2p'))
        self.thru5 = rf.Network(os.path.join(self.test_dir, 'deemb_thru5.s2p'))

        # for spot frequency checking
        self.raw5_1f = self.raw5['10GHz']
        self.thru5_1f = self.thru5['10GHz']

        # for impedance cancel cancel testing
        self.raw6 = rf.Network(os.path.join(self.test_dir, 'deemb_ind6.s2p'))
        self.thru6 = rf.Network(os.path.join(self.test_dir, 'deemb_thru6.s2p'))

        # for spot frequency checking
        self.raw6_1f = self.raw6['10GHz']
        self.thru6_1f = self.thru6['10GHz']

        # create de-embedding objects
        self.dm = rf.OpenShort(self.open, self.short)
        self.dm_os = rf.OpenShort(self.open_1f, self.short_1f)
        self.dm_o = rf.Open(self.open7_1f)
        self.dm_so = rf.ShortOpen(self.short2_1f, self.open2_1f)
        self.dm_s = rf.Short(self.short8_1f)
        self.dm_pi = rf.SplitPi(self.thru3_1f)
        self.dm_tee = rf.SplitTee(self.thru4_1f)
        self.dm_ac = rf.AdmittanceCancel(self.thru5_1f)
        self.dm_ic = rf.ImpedanceCancel(self.thru6_1f)

        # relative tolerance for comparisons
        self.rtol = 1e-6

    def test_freqmismatch(self):
        """
        Check that a warning is caught when networks are of different frequencies
        """
        with self.assertWarns(RuntimeWarning):
            rf.OpenShort(self.open, self.short[0:len(self.short)//2])

        with self.assertWarns(RuntimeWarning):
            self.dm.deembed(self.raw[0:len(self.raw)//2])

    def test_openshort(self):
        """
        After de-embedding, the network is a pure inductor of 1nH.
        Test that this is true at a spot frequency.
        """
        dut = self.dm_os.deembed(self.raw_1f)
        ind_calc = 1e9*np.imag(1/dut.y[0,0,0])/2/np.pi/dut.f
        self.assertTrue(np.isclose(ind_calc, 1, rtol=self.rtol))

    def test_open(self):
        """
        After de-embedding, the network is a pure inductor of 1nH.
        Test that this is true at a spot frequency.
        """
        dut = self.dm_o.deembed(self.raw7_1f)
        ind_calc = 1e9*np.imag(1/dut.y[0,0,0])/2/np.pi/dut.f
        self.assertTrue(np.isclose(ind_calc, 1, rtol=self.rtol))

    def test_shortopen(self):
        """
        After de-embedding, the network is a pure inductor of 1nH.
        Test that this is true at a spot frequency.
        """
        dut = self.dm_so.deembed(self.raw2_1f)
        ind_calc = 1e9*np.imag(1/dut.y[0,0,0])/2/np.pi/dut.f
        self.assertTrue(np.isclose(ind_calc, 1, rtol=self.rtol))

    def test_short(self):
        """
        After de-embedding, the network is a pure inductor of 1nH.
        Test that this is true at a spot frequency.
        """
        dut = self.dm_s.deembed(self.raw8_1f)
        ind_calc = 1e9*np.imag(1/dut.y[0,0,0])/2/np.pi/dut.f
        self.assertTrue(np.isclose(ind_calc, 1, rtol=self.rtol))

    def test_splitpi(self):
        """
        After de-embedding, the network is a pure inductor of 1nH.
        Test that this is true at a spot frequency.
        """
        dut = self.dm_pi.deembed(self.raw3_1f)
        ind_calc = 1e9*np.imag(1/dut.y[0,0,0])/2/np.pi/dut.f
        self.assertTrue(np.isclose(ind_calc, 1, rtol=self.rtol))

    def test_splittee(self):
        """
        After de-embedding, the network is a pure inductor of 1nH.
        Test that this is true at a spot frequency.
        """
        dut = self.dm_tee.deembed(self.raw4_1f)
        ind_calc = 1e9*np.imag(1/dut.y[0,0,0])/2/np.pi/dut.f
        self.assertTrue(np.isclose(ind_calc, 1, rtol=self.rtol))

    def test_admittance_cancel(self):
        """
        After de-embedding, the network is a pure inductor of 1nH.
        Test that this is true at a spot frequency.
        """
        dut = self.dm_ac.deembed(self.raw5_1f)
        ind_calc = 1e9*np.imag(1/dut.y[0,0,0])/2/np.pi/dut.f
        self.assertTrue(np.isclose(ind_calc, 1, rtol=self.rtol))

    def test_impedance_cancel(self):
        """
        After de-embedding, the network is a pure inductor of 1nH.
        Test that this is true at a spot frequency.
        """
        dut = self.dm_ic.deembed(self.raw6_1f)
        ind_calc = 1e9*np.imag(1/dut.y[0,0,0])/2/np.pi/dut.f
        self.assertTrue(np.isclose(ind_calc, 1, rtol=self.rtol))

    def test_IEEEP370_SE_NZC_2xThru(self):
        """
        Test test_IEEEP370_SE_NZC_2xThru.

        After de-embedding fixtures model from 2xtru, the network is a perfect
        thru.
        Test that this thru has S21 amplitude and phase smaller than a limit.
        """
        s2xthru = rf.Network(os.path.join(self.test_dir, 's2xthru.s2p'))
        dm_nzc = rf.IEEEP370_SE_NZC_2xThru(dummy_2xthru = s2xthru,
                                        name = '2xthru')
        residuals = dm_nzc.deembed(s2xthru)
        # insertion loss magnitude deviate from 1.0 from less than 0.1 dB
        il_mag = 20.*np.log10(np.abs(residuals.s[:, 1, 0] + 1e-12))
        self.assertTrue(np.max(np.abs(il_mag)) <= 0.1, 'residual IL magnitude')
        # insertion loss phase deviate from 0 degree from less than 1 degree
        il_phase = np.angle(residuals.s[:, 1, 0]) * 180/np.pi
        self.assertTrue(np.max(np.abs(il_phase)) <= 1.0, 'residual IL Phase')

    def test_IEEEP370_SE_NZC_2xThru_with_dc(self):
        """
        Test test_IEEEP370_SE_NZC_2xThru_with_dc.

        After de-embedding fixtures model from 2xtru, the network is a perfect
        thru.
        Test that this thru has S21 amplitude and phase smaller than a limit.
        """
        s2xthru = rf.Network(os.path.join(self.test_dir, 's2xthru.s2p'))
        # interpolate to dc
        s2xthru_dc = s2xthru.extrapolate_to_dc(kind='linear')
        with pytest.warns(RuntimeWarning, match="DC point detected"):
            dm_nzc = rf.IEEEP370_SE_NZC_2xThru(dummy_2xthru = s2xthru_dc,
                                        name = '2xthru')
        residuals = dm_nzc.deembed(s2xthru_dc)
        # insertion loss magnitude deviate from 1.0 from less than 0.1 dB
        il_mag = 20.*np.log10(np.abs(residuals.s[:, 1, 0] + 1e-12))
        self.assertTrue(np.max(np.abs(il_mag)) <= 0.1, 'residual IL magnitude')
        # insertion loss phase deviate from 0 degree from less than 1 degree
        il_phase = np.angle(residuals.s[:, 1, 0]) * 180/np.pi
        self.assertTrue(np.max(np.abs(il_phase)) <= 1.0, 'residual IL Phase')

    def test_IEEEP370_SE_NZC_2xThru_with_interpolation(self):
        """
        Test test_IEEEP370_SE_NZC_2xThru_with_interpolation.
        While interpolation should be avoided, it keep usefull in some cases,
        e.g. when the measurement frequency axis was not well formated but
        a result is still needed and some deviation acceptable.
        After de-embedding fixtures model from 2xtru, the network is a perfect
        thru.
        Test that this thru has S21 amplitude and phase smaller than a limit.
        """
        s2xthru = rf.Network(os.path.join(self.test_dir, 's2xthru.s2p'))
        # with non-hormonic sweep
        nonuniform_freq = rf.Frequency(s2xthru.f[0], s2xthru.f[-1],
                                       npoints=len(s2xthru)-10, unit='Hz')
        s2xthru_nu = s2xthru.interpolate(nonuniform_freq)
        with pytest.warns(RuntimeWarning, match="Non-uniform frequency vector detected"):
            dm_nzc_nu = rf.IEEEP370_SE_NZC_2xThru(dummy_2xthru = s2xthru_nu,
                                                name = '2xthru')
        residuals = dm_nzc_nu.deembed(s2xthru_nu)
        # insertion loss magnitude deviate from 1.0 from less than 0.1 dB
        il_mag = 20.*np.log10(np.abs(residuals.s[:, 1, 0] + 1e-12))
        self.assertTrue(np.max(np.abs(il_mag)) <= 0.1, 'residual IL magnitude')
        # insertion loss phase deviate from 0 degree from less than 1 degree
        il_phase = np.angle(residuals.s[:, 1, 0]) * 180/np.pi
        self.assertTrue(np.max(np.abs(il_phase)) <= 1.0, 'residual IL Phase')

    def test_IEEEP370_SE_ZC_2xThru(self):
        """
        Test test_IEEEP370_SE_ZC_2xThru.

        After de-embedding fixtures model from 2xtru, the network is a perfect
        thru.
        Test that this thru has S21 amplitude and phase smaller than a limit.
        """
        s2xthru = rf.Network(os.path.join(self.test_dir, 's2xthru.s2p'))
        fdf = rf.Network(os.path.join(self.test_dir, 'fdf.s2p'))
        dm_zc  = rf.IEEEP370_SE_ZC_2xThru(dummy_2xthru = s2xthru,
                                       dummy_fix_dut_fix = fdf,
                                       bandwidth_limit = 10e9,
                                       pullback1 = 0, pullback2 = 0,
                                       leadin = 0,
                                       NRP_enable = False,
                                       name = 'zc2xthru')
        residuals = dm_zc.deembed(s2xthru)
        # insertion loss magnitude deviate from 1.0 from less than 0.2 dB
        il_mag = 20.*np.log10(np.abs(residuals.s[:, 1, 0] + 1e-12))
        self.assertTrue(np.max(np.abs(il_mag)) <= 0.2, 'residual IL magnitude')
        # insertion loss phase deviate from 0 degree from less than 45 degree
        # too much tolerance here allowed as for now
        il_phase = np.angle(residuals.s[:, 1, 0]) * 180/np.pi
        self.assertTrue(np.max(np.abs(il_phase)) <= 2.0, 'residual IL Phase')

    def test_IEEEP370_SE_ZC_2xThru_with_dc(self):
        """
        Test test_IEEEP370_SE_ZC_2xThru_with_dc.

        After de-embedding fixtures model from 2xtru, the network is a perfect
        thru.
        Test that this thru has S21 amplitude and phase smaller than a limit.
        """
        s2xthru = rf.Network(os.path.join(self.test_dir, 's2xthru.s2p'))
        fdf = rf.Network(os.path.join(self.test_dir, 'fdf.s2p'))
        s2xthru_dc = s2xthru.extrapolate_to_dc(kind='linear')
        fdf_dc = fdf.extrapolate_to_dc(kind='linear')
        with pytest.warns(RuntimeWarning, match="DC point detected"):
            dm_zc  = rf.IEEEP370_SE_ZC_2xThru(dummy_2xthru = s2xthru_dc,
                                        dummy_fix_dut_fix = fdf_dc,
                                        bandwidth_limit = 10e9,
                                        pullback1 = 0, pullback2 = 0,
                                        leadin = 0,
                                        NRP_enable = False,
                                        name = 'zc2xthru')
        residuals = dm_zc.deembed(s2xthru_dc)
        # insertion loss magnitude deviate from 1.0 from less than 0.2 dB
        il_mag = 20.*np.log10(np.abs(residuals.s[:, 1, 0] + 1e-12))
        self.assertTrue(np.max(np.abs(il_mag)) <= 0.2, 'residual IL magnitude')
        # insertion loss phase deviate from 0 degree from less than 45 degree
        # too much tolerance here allowed as for now
        il_phase = np.angle(residuals.s[:, 1, 0]) * 180/np.pi
        self.assertTrue(np.max(np.abs(il_phase)) <= 2.0, 'residual IL Phase')

    def test_IEEEP370_SE_ZC_2xThru_with_interpolation(self):
        """
        Test test_IEEEP370_SE_ZC_2xThru_with_interpolation.
        While interpolation should be avoided, it keep usefull in some cases,
        e.g. when the measurement frequency axis was not well formated but
        a result is still needed and some deviation acceptable.
        After de-embedding fixtures model from 2xtru, the network is a perfect
        thru.
        Test that this thru has S21 amplitude and phase smaller than a limit.
        """
        s2xthru = rf.Network(os.path.join(self.test_dir, 's2xthru.s2p'))
        fdf = rf.Network(os.path.join(self.test_dir, 'fdf.s2p'))
        # with non-hormonic sweep
        nonuniform_freq = rf.Frequency(s2xthru.f[0], s2xthru.f[-1],
                                       npoints=len(s2xthru)-10, unit='Hz')
        s2xthru_nu = s2xthru.interpolate(nonuniform_freq)
        fdf_nu = fdf.interpolate(nonuniform_freq)
        with pytest.warns(RuntimeWarning, match="Non-uniform frequency vector detected"):
            dm_zc_nu  = rf.IEEEP370_SE_ZC_2xThru(dummy_2xthru = s2xthru_nu,
                                        dummy_fix_dut_fix = fdf_nu,
                                        bandwidth_limit = 10e9,
                                        pullback1 = 0, pullback2 = 0,
                                        leadin = 0,
                                        NRP_enable = False,
                                        name = 'zc2xthru')
        residuals = dm_zc_nu.deembed(s2xthru_nu)
        # insertion loss magnitude deviate from 1.0 from less than 0.2 dB
        il_mag = 20.*np.log10(np.abs(residuals.s[:, 1, 0] + 1e-12))
        self.assertTrue(np.max(np.abs(il_mag)) <= 0.2, 'residual IL magnitude')
        # insertion loss phase deviate from 0 degree from less than 45 degree
        # too much tolerance here allowed as for now
        il_phase = np.angle(residuals.s[:, 1, 0]) * 180/np.pi
        self.assertTrue(np.max(np.abs(il_phase)) <= 2.0, 'residual IL Phase')

    def test_IEEEP370_SE_NZC_2xThru_no_180deg_phase_jumps(self):
        """
        Test test_IEEEP370_SE_NZC_2xThru_no_180deg_phase_jumps.

        After de-embedding fixtures model from 2xtru, the network is a perfect
        thru.
        If there is noise on the 2xthru phase, incorrect computation of e01 and
        e10 could lead to 180° jumps into the side models s21 phase.
        Check this not happens.
        """
        s2xthru = rf.Network(os.path.join(self.test_dir, 's2xthru.s2p'))
        # with phase noise
        # add enough phase noise to trigger phase jumps with original
        # implementation, but small enough to keep within 1° limit line
        s2xthru_pn = s2xthru.copy()
        s2xthru_pn.add_noise_polar(0.0002, 0.2)
        dm_nzc_pn = rf.IEEEP370_SE_NZC_2xThru(dummy_2xthru = s2xthru_pn,
                                        name = '2xthru')
        residuals = dm_nzc_pn.deembed(s2xthru_pn)
        # insertion loss magnitude deviate from 1.0 from less than 0.1 dB
        il_mag = 20.*np.log10(np.abs(residuals.s[:, 1, 0] + 1e-12))
        self.assertTrue(np.max(np.abs(il_mag)) <= 0.1, 'residual IL magnitude')
        # insertion loss phase deviate from 0 degree from less than 1 degree
        il_phase = np.angle(residuals.s[:, 1, 0]) * 180/np.pi
        self.assertTrue(np.max(np.abs(il_phase)) <= 1.0, 'residual IL Phase')

    def test_IEEEP370_check_fd_td_se_quality(self):
        """
        Test test_IEEEP370_check_fd_se_quality.

        Check for passivity, reciprocity, and causality in the frequency
        and in the time domains.
        """
        freq = rf.Frequency(10e-3, 10, 1000, 'GHz')
        W   = 3.00e-3
        H   = 1.55e-3
        T   = 50e-6
        ep_r = 4.459
        tanD = 0.0183
        f_epr_tand = 1e9

        beta  = 2 * pi * freq.f / c # propagation in air
        # 50 ohm air line
        m50 = rf.media.DefinedGammaZ0(frequency = freq, z0_port = 50, gamma = 1j * beta)

        # microstrip segments
        with pytest.warns(RuntimeWarning, match = r"^Conductor"):
            m = MLine(frequency=freq, z0_port=50, w=W, h=H, t=T,
                    ep_r=ep_r, mu_r=1, rho=1.712e-8, tand=tanD, rough=0.15e-6,
                    f_low=1e3, f_high=1e12, f_epr_tand=f_epr_tand,
                    diel='djordjevicsvensson', disp='kirschningjansen')

        thru  = m.line(0.050, 'm', z0 = 52.5)
        thru.name = "thru"
        # perfect data
        fd_qm = rf.IEEEP370_FD_QM()
        td_qm = rf.IEEEP370_TD_QM(1e9, # bps
                                  32,  # samples per UI
                                  0.4, # rise time per UI
                                  1,   # gaussian pulse
                                  2,   # zero extrapolation
                                  verbose = False)
        qm = fd_qm.check_se_quality(thru)
        self.assertTrue(qm['passivity']['value'] > 99.9 and
                        qm['reciprocity']['value'] > 99.9 and
                        qm['causality']['value'] > 99.9,
                        'FD quality perfect thru')
        # passivity violation (reference value from Matlab R2024a)
        thru_non_passive = thru.copy()
        thru_non_passive.s[:, 1, 0] = 1.137 * thru_non_passive.s[:, 1, 0]
        thru_non_passive.s[:, 0, 1] = 1.137 * thru_non_passive.s[:, 0, 1]
        qm = fd_qm.check_se_quality(thru_non_passive)
        self.assertTrue(np.round(qm['passivity']['value'], 4) == 49.7701,
                        'FD quality passivity violation')
        qm = td_qm.check_se_quality(thru_non_passive)
        self.assertTrue(np.round(qm['passivity']['value'], 4) == 71.3500,
                        'TD quality passivity violation')
        # reciprocity violation (reference value from Matlab R2024a)
        thru_non_reciprocal = thru.copy()
        thru_non_reciprocal.s[:, 1, 0] = 0.945 * thru_non_reciprocal.s[:, 0, 1]
        qm = fd_qm.check_se_quality(thru_non_reciprocal)
        self.assertTrue(np.round(qm['reciprocity']['value'], 4) == 49.6120,
                        'FD quality reciprocity violation')
        qm = td_qm.check_se_quality(thru_non_reciprocal)
        self.assertTrue(np.round(qm['reciprocity']['value'], 4) == 28.4500,
                        'TD quality reciprocity violation')
        # causality violation (reference value from Matlab R2024a)
        half_fd  = m50.line(0.0445, 'm', z0 = 52)
        thru_non_causal_fd = half_fd.inv ** thru ** half_fd.inv
        thru_non_causal_fd.name = 'thru'
        qm = fd_qm.check_se_quality(thru_non_causal_fd)
        self.assertTrue(np.round(qm['causality']['value'], 4) == 48.2637,
                        'FD quality causality violation')
        half  = m50.line(0.18, 'm', z0 = 52)
        thru_non_causal = half.inv ** thru ** half.inv
        thru_non_causal.name = 'thru'
        qm = td_qm.check_se_quality(thru_non_causal)
        self.assertTrue(np.round(qm['causality']['value'], 4) == 10.0500,
                        'TD quality causality violation')




"""
.. module:: skrf.calibration.calibrationSet
================================================================
calibrationSet (:mod:`skrf.calibration.calibrationSet`)
================================================================


Contains the CalibrationSet class, and supporting functions

CalibrationSet Class
===============

.. autosummary::
   :toctree: generated/

   CalibrationSet

"""
from itertools import product

from ..networkSet import NetworkSet
from .calibration import Calibration


def cartesian_product(ideals, measured_sets, *args, **kwargs):
    """
    """
    measured_lists = product(*[k[:] for k in measured_sets])
    return [Calibration(ideals = ideals, measured = measured,
        **kwargs) for measured in measured_lists ]

def dot_product(ideals, measured_sets, *args, **kwargs):
    """
    """
    for measured_set in measured_sets:
        if len(measured_set) != len(measured_sets[0]):
            raise(IndexError('all measured NetworkSets must have same length for dot product combinatoric function'))

    cal_list = []
    for k in list(range(len(measured_sets[0]))):
        measured = [measured_set[k] for measured_set in measured_sets]
        cal_list.append(
            Calibration(ideals=ideals, measured= measured,
            **kwargs)
            )

    return cal_list

class CalibrationSet:
    """
    A set of Calibrations

    This is designed to support experimental uncertainty analysis [1]_.

    References
    -----------

    .. [1] A. Arsenovic, L. Chen, M. F. Bauwens, H. Li, N. S. Barker, and R. M. Weikle,
    "An Experimental Technique for Calibration Uncertainty Analysis,"
    IEEE Transactions on Microwave Theory and Techniques, vol. 61, no. 1, pp. 263-269, 2013.

    """

    def __init__(self, cal_class, ideals, measured_sets,*args, **kwargs):
        r"""
        Parameters
        ----------
        cal_class : a Calibration class
            this is the class of calibration to use on the set. This
            argument is the actual class itself like OnePort, TRL, SOLT, etc

        ideals : list of Networks

        measured_set :  list of NetworkSets, or list of lists
            each element in this list should be a corresponding measured
            set to the ideals element of the same index. The sets
            themselves  can be anything list-like

        \*args\**kargs :
            passed to self.run(),

        """
        self.cal_class = cal_class
        self.ideals = ideals
        self.measured_sets = measured_sets
        self.args = args
        self.kwargs = kwargs
        self.run(*args, **kwargs)

    def __getitem__(self, key):
        return self.cal_list[key]

    def apply_cal(self, raw_ntwk, *args, **kwargs):
        """
        """
        return NetworkSet([k.apply_cal(raw_ntwk) for k in self.cal_list],
            *args, **kwargs)

    def plot_uncertainty_per_standard(self):
        """
        """
        self.dankness('std_s','plot_s_mag')

    def dankness(self, prop, func, *args, **kwargs):
        """
        """
        try:
            [getattr(getattr(k, prop), func) \
                (*args, **kwargs) for k in self.measured_sets]
        except (TypeError):
            return [getattr(getattr(k, prop), func) \
                for k in self.measured_sets]

    def run(self):
        NotImplementedError('SubClass must implement this')

    @property
    def corrected_sets(self):
        """
        The set of corrected networks, each is corrected by its corresponding
        element in the cal_list
        """
        n_meas = len(self.cal_list[0].measured)
        mat = [k.caled_ntwks for k in self.cal_list]
        return [NetworkSet([k[l] for k in mat]) for l in range(n_meas)]



class Dot(CalibrationSet):

    def run(self, *args, **kwargs):
        ideals = self.ideals
        measured_sets = self.measured_sets
        if len(set(map(len, measured_sets))) !=1:
            raise(IndexError('all measured NetworkSets must have same length for dot product combinatoric function'))

        self.cal_list = []
        for k in range(len(measured_sets[0])):
            measured = [measured_set[k] for measured_set in measured_sets]
            cal = self.cal_class(ideals=ideals, measured= measured,
                                 **kwargs)
            self.cal_list.append(cal)



"""
freespace (:mod:`skrf.media.freespace`)
========================================

A plane-wave (TEM Mode) in Freespace.

Represents a plane-wave in a homogeneous freespace, defined by
the space's relative permittivity and relative permeability.

.. autosummary::
    :toctree: generated/

    Freespace


"""
from __future__ import annotations

from typing import TYPE_CHECKING

from numpy import ones, real, sqrt
from scipy.constants import epsilon_0, mu_0

from ..constants import NumberLike
from ..data import materials
from .media import Media

if TYPE_CHECKING:
    from ..frequency import Frequency


class Freespace(Media):
    r"""
    A plane-wave (TEM Mode) in Freespace.

    A Freespace media can be constructed in two ways:

    * from complex, relative permativity and permeability OR
    * from real relative permativity and permeability with loss tangents.

    There is also a method to initialize from a
    existing distributed circuit, appropriately named
    :func:`Freespace.from_distributed_circuit`


    Parameters
    ----------
    frequency : :class:`~skrf.frequency.Frequency` object
        frequency band of this transmission line medium
    z0_port : number, array-like, or None
        `z0_port` is the port impedance for networks generated by the media.
        If `z0_port` is not None, the networks generated by the media are
        renormalized (or in other words embedded) from the characteristic
        impedance z0 of the media to `z0_port`.
        Else if `z0_port` is None, the networks port impedances will be the raw
        characteristic impedance z0 of the media.
        (Default is None)
    z0_override : number, array-like, or None
        `z0_override` override the characteristic impedance for the media.
        If `z0_override` is not None, the networks generated by the media have
        their characteristic impedance `z0` overrided by `z0_override`.
        (Default is None)
    z0 : number, array-like, or None
        deprecated parameter, alias to `z0_override` if `z0_override` is None.
        Emmit a deprecation warning.
    ep_r : number, array-like
        complex relative permittivity. negative imaginary is lossy.
    mu_r : number, array-like
        complex relative permeability. negative imaginary is lossy.
    ep_loss_tan : None, number, array-like
        electric loss tangent (of the permativity).
        If not None, imag(ep_r) is ignored.
    mu_loss_tan : None, number, array-like
        magnetic loss tangent (of the permeability).
        If not None, imag(mu_r) is ignored.
    rho : number, array-like, string or None
        resistivity (ohm-m) of the conductor walls. If array-like
        must be same length as frequency. if str, it must be a key in
        :data:`skrf.data.materials`.
        Default is None (lossless).
    \*args, \*\*kwargs : arguments and keyword arguments


    Examples
    --------
    >>> from skrf.media.freespace import Freespace
    >>> from skrf.frequency import Frequency
    >>> f = Frequency(75,110,101,'ghz')
    >>> Freespace(frequency=f, ep_r=11.9)
    >>> Freespace(frequency=f, ep_r=11.9-1.1j)
    >>> Freespace(frequency=f, ep_r=11.9, ep_loss_tan=.1)
    >>> Freespace(frequency=f, ep_r=11.9-1.1j, mu_r = 1.1-.1j)

    """

    def __init__(self, frequency: Frequency | None = None,
                 z0_port: NumberLike | None = None,
                 z0_override: NumberLike | None = None,
                 z0: NumberLike | None = None,
                 ep_r: NumberLike = 1+0j, mu_r: NumberLike = 1+0j,
                 ep_loss_tan: NumberLike | None = None,
                 mu_loss_tan: NumberLike | None = None,
                 rho: NumberLike | str | None = None,
                 *args, **kwargs):
        Media.__init__(self, frequency = frequency,
                       z0_port = z0_port, z0_override = z0_override, z0 = z0)
        self.ep_r = ep_r
        self.mu_r = mu_r
        self.rho = rho

        self.ep_loss_tan = ep_loss_tan
        self.mu_loss_tan = mu_loss_tan

    def __str__(self) -> str:
        f = self.frequency
        output = 'Freespace  Media.  %i-%i %s.  %i points'%\
                (f.f_scaled[0], f.f_scaled[-1], f.unit, f.npoints)
        return output

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def ep(self) -> NumberLike:
        r"""
        Complex dielectric permittivity.

        If :math:`\tan\delta_e` is not defined:

        .. math::

            \varepsilon = \varepsilon_0 \varepsilon_r

        otherwise,

        .. math::

            \varepsilon = \varepsilon_0 \Re[\varepsilon_r] (1 - j\tan\delta_e)

        where :math:`\tan\delta_e` is the electric loss tangent.

        Returns
        -------
        ep : number or array-like
            Complex dielectric permittivity in F/m.

        """

        if self.ep_loss_tan is not None:
            ep_r = real(self.ep_r)*(1 - 1j*self.ep_loss_tan)
        else:
            ep_r = self.ep_r
        return ep_r*epsilon_0

    @property
    def mu(self) -> NumberLike:
        r"""
        Complex dielectric permeability.

        If :math:`\tan\delta_m` is not defined:

        .. math::

            \mu = \mu_0 \mu_r

        otherwise,

        .. math::

            \mu = \mu_0 \Re[\mu_r] (1 - j\tan\delta_m)

        where :math:`\tan\delta_m` is the magnetic loss tangent.

        Returns
        -------
        mu : number
            Complex permeability in H/m.
        """
        if self.mu_loss_tan is not None:
            mu_r = real(self.mu_r)*(1 -1j*self.mu_loss_tan)
        else:
            mu_r = self.mu_r
        return mu_r*mu_0


    @classmethod
    def from_distributed_circuit(cls, dc, *args, **kwargs) -> Media:
        r"""
        Initialize a freespace from :class:`~skrf.media.distributedCircuit.DistributedCircuit`.

        Parameters
        ----------
        dc: :class:`~skrf.media.distributedCircuit.DistributedCircuit`
            a DistributedCircuit object
        \*args, \*\*kwargs :
            passed to `Freespace.__init__

        Notes
        -----
        Here are the details::

            w = dc.frequency.w
            z= dc.Z/(w*mu_0)
            y= dc.Y/(w*epsilon_0)
            ep_r = -1j*y
            mu_r = -1j*z

        See Also
        --------
        skrf.media.distributedCircuit.DistributedCircuit

        """
        w = dc.frequency.w
        z= dc.Z/(w*mu_0)
        y= dc.Y/(w*epsilon_0)


        kw={}
        kw['ep_r'] = -1j*y
        kw['mu_r'] = -1j*z

        kwargs.update(kw)
        return cls(frequency=dc.frequency, **kwargs)

    @property
    def rho(self) -> NumberLike:
        """
        Conductivity in ohm*m.

        Parameters
        ----------
        val : float, array-like or str
            the resistivity in ohm*m. If array-like must be same length
            as self.frequency. if str, it must be a key in
            :data:`~skrf.data.materials`.

        Examples
        --------
        >>> wg.rho = 2.8e-8
        >>> wg.rho = 2.8e-8 * ones(len(wg.frequency))
        >>> wg.rho = 'al'
        >>> wg.rho = 'aluminum'
        """
        return self._rho

    @rho.setter
    def rho(self, val):
        if isinstance(val, str):
            self._rho = materials[val.lower()]['resistivity(ohm*m)']
        else:
            self._rho=val

    @property
    def ep_with_rho(self) -> NumberLike:
        r"""
        Complex permittivity with resistivity absorbed into its imaginary component.

        .. math::

            \varepsilon - j \frac{1}{\rho\omega}

        See Also
        --------
        rho
        ep
        """
        if self.rho is not None:
            return self.ep -1j/(self.rho*self.frequency.w)
        else:
            return self.ep

    @property
    def gamma(self) -> NumberLike:
        r"""
        Propagation Constant, :math:`\gamma`.

        Defined as,

        .. math::

                \gamma =  \sqrt{ Z^{'}  Y^{'}}

        Returns
        -------
        gamma : np.ndarray
            Propagation Constant,

        Note
        ----
        The components of propagation constant are interpreted as follows:

        * positive real(gamma) = attenuation
        * positive imag(gamma) = forward propagation
        """
        ep = self.ep_with_rho
        return 1j*self.frequency.w * sqrt(ep*self.mu)

    @property
    def z0_characteristic(self) -> NumberLike:
        r"""
        Characteristic Impedance, :math:`z_0`.

        .. math::

                Z_0 = \sqrt{ \frac{Z^{'}}{Y^{'}}}

        Returns
        -------
        z0_characteristic : np.ndarray
            Characteristic Impedance in units of ohms
        """
        ep = self.ep_with_rho
        return sqrt(self.mu/ep)*ones(len(self))

    def plot_ep(self):
        """
        Plot the real and imaginary part of the complex permittivity.
        """
        self.plot(self.ep_r.real, label=r'ep_r real')
        self.plot(self.ep_r.imag, label=r'ep_r imag')

    def plot_mu(self):
        """
        Plot the real and imaginary part of the complex permeability.
        """
        self.plot(self.mu_r.real, label=r'mu_r real')
        self.plot(self.mu_r.imag, label=r'mu_r imag')

    def plot_ep_mu(self):
        """
        Plot the real and imaginary part of the complex permittivity with resistivity.
        """
        self.plot_ep()
        self.plot_mu()



"""
skrf is an object-oriented approach to microwave engineering,
implemented in Python.
"""

__version__ = '1.5.0'
## Import all  module names for coherent reference of name-space
#import io


from . import (
    calibration,
    circuit,
    constants,
    frequency,
    instances,
    io,
    mathFunctions,
    media,
    network,
    networkSet,
    qfactor,
    taper,
    tlineFunctions,
    util,
    vectorFitting,
)
from .calibration import *
from .circuit import *
from .constants import *

# Import contents into current namespace for ease of calling
from .frequency import *
from .instances import *
from .io import *
from .mathFunctions import *
from .network import *
from .networkSet import *
from .qfactor import *
from .taper import *
from .tlineFunctions import *
from .util import *
from .vectorFitting import *

# Try to import vi, but if except if pyvisa not installed
try:
    import vi
    from vi import *
except ImportError:
    pass

# try to import data but if it fails whatever. it fails if some pickles
# dont unpickle. but its not important
try:
    from . import data
except Exception:
    pass

def __getattr__(name: str):
    return getattr(instances._instances, name)

## built-in imports
from copy import deepcopy as copy

## Shorthand Names
F = Frequency
N = Network
NS = NetworkSet
C = Circuit
lat = load_all_touchstones
# saf  = save_all_figs
saf = None
stylely = None


def setup_pylab() -> bool:
    try:
        import matplotlib
    except ImportError:
        print("matplotlib not found while setting up plotting")
        return False

    from . import plotting

    global saf, stylely
    saf = plotting.save_all_figs
    stylely = plotting.stylely
    return True


def setup_plotting():
    plotting_environment = os.environ.get('SKRF_PLOT_ENV', "pylab").lower()

    if plotting_environment == "pylab":
        setup_pylab()
    elif plotting_environment == "pylab-skrf-style":
        if setup_pylab():
            stylely()
    # elif some different plotting environment
        # set that up

plotting_available = setup_plotting()
